@@ -33,6 +33,8 @@ SERVICE_PROVIDERS = {
     "MISTRAL": "Mistral",
     "EMBEDCHAIN": "Embedchain",
     "AUTOGEN": "Autogen",
+    "XAI": "XAI",
+    "AWS_BEDROCK": "AWS Bedrock",
 }
 
 LANGTRACE_ADDITIONAL_SPAN_ATTRIBUTES_KEY = "langtrace_additional_attributes"
