@@ -1,21 +1,5 @@
-"""
-Copyright (c) 2024 Scale3 Labs
-
-Licensed under the Apache License, Version 2.0 (the "License");
-you may not use this file except in compliance with the License.
-You may obtain a copy of the License at
-
-    http://www.apache.org/licenses/LICENSE-2.0
-
-Unless required by applicable law or agreed to in writing, software
-distributed under the License is distributed on an "AS IS" BASIS,
-WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-See the License for the specific language governing permissions and
-limitations under the License.
-"""
-
 import json
-
+from typing import Any, Dict, List, Optional, Callable, Awaitable, Union
 from langtrace.trace_attributes import (
     LLMSpanAttributes,
     SpanAttributes,
@@ -23,7 +7,7 @@ from langtrace.trace_attributes import (
 from langtrace_python_sdk.utils import set_span_attribute
 from langtrace_python_sdk.utils.silently_fail import silently_fail
 from opentelemetry import trace
-from opentelemetry.trace import SpanKind
+from opentelemetry.trace import SpanKind, Tracer, Span
 from opentelemetry.trace.status import Status, StatusCode
 from opentelemetry.trace.propagation import set_span_in_context
 from langtrace_python_sdk.constants.instrumentation.common import (
@@ -37,34 +21,55 @@ from langtrace_python_sdk.utils.llm import (
     get_langtrace_attributes,
     get_llm_request_attributes,
     get_llm_url,
+    get_span_name,
     get_tool_calls,
     is_streaming,
     set_event_completion,
     StreamWrapper,
     set_span_attributes,
 )
-from openai._types import NOT_GIVEN
+from langtrace_python_sdk.types import NOT_GIVEN
+
+from langtrace_python_sdk.instrumentation.openai.types import (
+    ImagesGenerateKwargs,
+    ChatCompletionsCreateKwargs,
+    EmbeddingsCreateKwargs,
+    ImagesEditKwargs,
+    ResultType,
+    ContentItem,
+)
 
 
-def images_generate(original_method, version, tracer):
+def filter_valid_attributes(attributes):
+    """Filter attributes where value is not None, not an empty string, and not openai.NOT_GIVEN."""
+    return {
+        key: value
+        for key, value in attributes.items()
+        if value is not None and value != NOT_GIVEN and value != ""
+    }
+
+
+def images_generate(version: str, tracer: Tracer) -> Callable:
     """
     Wrap the `generate` method of the `Images` class to trace it.
     """
 
-    def traced_method(wrapped, instance, args, kwargs):
+    def traced_method(
+        wrapped: Callable, instance: Any, args: List[Any], kwargs: ImagesGenerateKwargs
+    ) -> Any:
         service_provider = SERVICE_PROVIDERS["OPENAI"]
         span_attributes = {
             **get_langtrace_attributes(version, service_provider, vendor_type="llm"),
             **get_llm_request_attributes(kwargs, operation_name="images_generate"),
             **get_llm_url(instance),
             SpanAttributes.LLM_PATH: APIS["IMAGES_GENERATION"]["ENDPOINT"],
-            **get_extra_attributes(),
+            **get_extra_attributes(),  # type: ignore
         }
 
-        attributes = LLMSpanAttributes(**span_attributes)
+        attributes = LLMSpanAttributes(**filter_valid_attributes(span_attributes))
 
         with tracer.start_as_current_span(
-            APIS["IMAGES_GENERATION"]["METHOD"],
+            name=get_span_name(APIS["IMAGES_GENERATION"]["METHOD"]),
             kind=SpanKind.CLIENT,
             context=set_span_in_context(trace.get_current_span()),
         ) as span:
@@ -73,21 +78,17 @@ def images_generate(original_method, version, tracer):
                 # Attempt to call the original method
                 result = wrapped(*args, **kwargs)
                 if not is_streaming(kwargs):
-                    data = (
+                    data: Optional[ContentItem] = (
                         result.data[0]
                         if hasattr(result, "data") and len(result.data) > 0
-                        else {}
+                        else None
                     )
                     response = [
                         {
                             "role": "assistant",
                             "content": {
-                                "url": data.url if hasattr(data, "url") else "",
-                                "revised_prompt": (
-                                    data.revised_prompt
-                                    if hasattr(data, "revised_prompt")
-                                    else ""
-                                ),
+                                "url": getattr(data, "url", ""),
+                                "revised_prompt": getattr(data, "revised_prompt", ""),
                             },
                         }
                     ]
@@ -108,12 +109,14 @@ def images_generate(original_method, version, tracer):
     return traced_method
 
 
-def async_images_generate(original_method, version, tracer):
+def async_images_generate(version: str, tracer: Tracer) -> Callable:
     """
     Wrap the `generate` method of the `Images` class to trace it.
     """
 
-    async def traced_method(wrapped, instance, args, kwargs):
+    async def traced_method(
+        wrapped: Callable, instance: Any, args: List[Any], kwargs: ImagesGenerateKwargs
+    ) -> Awaitable[Any]:
         service_provider = SERVICE_PROVIDERS["OPENAI"]
 
         span_attributes = {
@@ -121,13 +124,13 @@ def async_images_generate(original_method, version, tracer):
             **get_llm_request_attributes(kwargs, operation_name="images_generate"),
             **get_llm_url(instance),
             SpanAttributes.LLM_PATH: APIS["IMAGES_GENERATION"]["ENDPOINT"],
-            **get_extra_attributes(),
+            **get_extra_attributes(),  # type: ignore
         }
 
-        attributes = LLMSpanAttributes(**span_attributes)
+        attributes = LLMSpanAttributes(**filter_valid_attributes(span_attributes))
 
         with tracer.start_as_current_span(
-            APIS["IMAGES_GENERATION"]["METHOD"],
+            name=get_span_name(APIS["IMAGES_GENERATION"]["METHOD"]),
             kind=SpanKind.CLIENT,
             context=set_span_in_context(trace.get_current_span()),
         ) as span:
@@ -136,21 +139,17 @@ def async_images_generate(original_method, version, tracer):
                 # Attempt to call the original method
                 result = await wrapped(*args, **kwargs)
                 if not is_streaming(kwargs):
-                    data = (
+                    data: Optional[ContentItem] = (
                         result.data[0]
                         if hasattr(result, "data") and len(result.data) > 0
-                        else {}
+                        else None
                     )
                     response = [
                         {
                             "role": "assistant",
                             "content": {
-                                "url": data.url if hasattr(data, "url") else "",
-                                "revised_prompt": (
-                                    data.revised_prompt
-                                    if hasattr(data, "revised_prompt")
-                                    else ""
-                                ),
+                                "url": getattr(data, "url", ""),
+                                "revised_prompt": getattr(data, "revised_prompt", ""),
                             },
                         }
                     ]
@@ -171,12 +170,14 @@ def async_images_generate(original_method, version, tracer):
     return traced_method
 
 
-def images_edit(original_method, version, tracer):
+def images_edit(version: str, tracer: Tracer) -> Callable:
     """
     Wrap the `edit` method of the `Images` class to trace it.
     """
 
-    def traced_method(wrapped, instance, args, kwargs):
+    def traced_method(
+        wrapped: Callable, instance: Any, args: List[Any], kwargs: ImagesEditKwargs
+    ) -> Any:
         service_provider = SERVICE_PROVIDERS["OPENAI"]
 
         span_attributes = {
@@ -186,13 +187,13 @@ def images_edit(original_method, version, tracer):
             SpanAttributes.LLM_PATH: APIS["IMAGES_EDIT"]["ENDPOINT"],
             SpanAttributes.LLM_RESPONSE_FORMAT: kwargs.get("response_format"),
             SpanAttributes.LLM_IMAGE_SIZE: kwargs.get("size"),
-            **get_extra_attributes(),
+            **get_extra_attributes(),  # type: ignore
         }
 
-        attributes = LLMSpanAttributes(**span_attributes)
+        attributes = LLMSpanAttributes(**filter_valid_attributes(span_attributes))
 
         with tracer.start_as_current_span(
-            APIS["IMAGES_EDIT"]["METHOD"],
+            name=APIS["IMAGES_EDIT"]["METHOD"],
             kind=SpanKind.CLIENT,
             context=set_span_in_context(trace.get_current_span()),
         ) as span:
@@ -232,10 +233,15 @@ def images_edit(original_method, version, tracer):
     return traced_method
 
 
-def chat_completions_create(original_method, version, tracer):
+def chat_completions_create(version: str, tracer: Tracer) -> Callable:
     """Wrap the `create` method of the `ChatCompletion` class to trace it."""
 
-    def traced_method(wrapped, instance, args, kwargs):
+    def traced_method(
+        wrapped: Callable,
+        instance: Any,
+        args: List[Any],
+        kwargs: ChatCompletionsCreateKwargs,
+    ) -> Any:
         service_provider = SERVICE_PROVIDERS["OPENAI"]
         if "perplexity" in get_base_url(instance):
             service_provider = SERVICE_PROVIDERS["PPLX"]
@@ -243,6 +249,8 @@ def chat_completions_create(original_method, version, tracer):
             service_provider = SERVICE_PROVIDERS["AZURE"]
         elif "groq" in get_base_url(instance):
             service_provider = SERVICE_PROVIDERS["GROQ"]
+        elif "x.ai" in get_base_url(instance):
+            service_provider = SERVICE_PROVIDERS["XAI"]
         llm_prompts = []
         for item in kwargs.get("messages", []):
             tools = get_tool_calls(item)
@@ -250,21 +258,13 @@ def chat_completions_create(original_method, version, tracer):
                 tool_calls = []
                 for tool_call in tools:
                     tool_call_dict = {
-                        "id": tool_call.id if hasattr(tool_call, "id") else "",
-                        "type": tool_call.type if hasattr(tool_call, "type") else "",
+                        "id": getattr(tool_call, "id", ""),
+                        "type": getattr(tool_call, "type", ""),
                     }
                     if hasattr(tool_call, "function"):
                         tool_call_dict["function"] = {
-                            "name": (
-                                tool_call.function.name
-                                if hasattr(tool_call.function, "name")
-                                else ""
-                            ),
-                            "arguments": (
-                                tool_call.function.arguments
-                                if hasattr(tool_call.function, "arguments")
-                                else ""
-                            ),
+                            "name": getattr(tool_call.function, "name", ""),
+                            "arguments": getattr(tool_call.function, "arguments", ""),
                         }
                     tool_calls.append(tool_call_dict)
                 llm_prompts.append(tool_calls)
@@ -276,13 +276,13 @@ def chat_completions_create(original_method, version, tracer):
             **get_llm_request_attributes(kwargs, prompts=llm_prompts),
             **get_llm_url(instance),
             SpanAttributes.LLM_PATH: APIS["CHAT_COMPLETION"]["ENDPOINT"],
-            **get_extra_attributes(),
+            **get_extra_attributes(),  # type: ignore
         }
 
-        attributes = LLMSpanAttributes(**span_attributes)
+        attributes = LLMSpanAttributes(**filter_valid_attributes(span_attributes))
 
         span = tracer.start_span(
-            APIS["CHAT_COMPLETION"]["METHOD"],
+            name=get_span_name(APIS["CHAT_COMPLETION"]["METHOD"]),
             kind=SpanKind.CLIENT,
             context=set_span_in_context(trace.get_current_span()),
         )
@@ -296,12 +296,9 @@ def chat_completions_create(original_method, version, tracer):
                     prompt_tokens += calculate_prompt_tokens(
                         json.dumps(str(message)), kwargs.get("model")
                     )
-
-                if (
-                    kwargs.get("functions") is not None
-                    and kwargs.get("functions") != NOT_GIVEN
-                ):
-                    for function in kwargs.get("functions"):
+                functions = kwargs.get("functions")
+                if functions is not None and functions != NOT_GIVEN:
+                    for function in functions:
                         prompt_tokens += calculate_prompt_tokens(
                             json.dumps(function), kwargs.get("model")
                         )
@@ -314,7 +311,7 @@ def chat_completions_create(original_method, version, tracer):
                     tool_calls=kwargs.get("tools") is not None,
                 )
             else:
-                _set_response_attributes(span, kwargs, result)
+                _set_response_attributes(span, result)
                 span.set_status(StatusCode.OK)
                 span.end()
                 return result
@@ -328,15 +325,22 @@ def chat_completions_create(original_method, version, tracer):
     return traced_method
 
 
-def async_chat_completions_create(original_method, version, tracer):
+def async_chat_completions_create(version: str, tracer: Tracer) -> Callable:
     """Wrap the `create` method of the `ChatCompletion` class to trace it."""
 
-    async def traced_method(wrapped, instance, args, kwargs):
+    async def traced_method(
+        wrapped: Callable,
+        instance: Any,
+        args: List[Any],
+        kwargs: ChatCompletionsCreateKwargs,
+    ) -> Awaitable[Any]:
         service_provider = SERVICE_PROVIDERS["OPENAI"]
         if "perplexity" in get_base_url(instance):
             service_provider = SERVICE_PROVIDERS["PPLX"]
         elif "azure" in get_base_url(instance):
             service_provider = SERVICE_PROVIDERS["AZURE"]
+        elif "x.ai" in get_base_url(instance):
+            service_provider = SERVICE_PROVIDERS["XAI"]
         llm_prompts = []
         for item in kwargs.get("messages", []):
             tools = get_tool_calls(item)
@@ -344,21 +348,13 @@ def async_chat_completions_create(original_method, version, tracer):
                 tool_calls = []
                 for tool_call in tools:
                     tool_call_dict = {
-                        "id": tool_call.id if hasattr(tool_call, "id") else "",
-                        "type": tool_call.type if hasattr(tool_call, "type") else "",
+                        "id": getattr(tool_call, "id", ""),
+                        "type": getattr(tool_call, "type", ""),
                     }
                     if hasattr(tool_call, "function"):
                         tool_call_dict["function"] = {
-                            "name": (
-                                tool_call.function.name
-                                if hasattr(tool_call.function, "name")
-                                else ""
-                            ),
-                            "arguments": (
-                                tool_call.function.arguments
-                                if hasattr(tool_call.function, "arguments")
-                                else ""
-                            ),
+                            "name": getattr(tool_call.function, "name", ""),
+                            "arguments": getattr(tool_call.function, "arguments", ""),
                         }
                     tool_calls.append(json.dumps(tool_call_dict))
                 llm_prompts.append(tool_calls)
@@ -370,13 +366,13 @@ def async_chat_completions_create(original_method, version, tracer):
             **get_llm_request_attributes(kwargs, prompts=llm_prompts),
             **get_llm_url(instance),
             SpanAttributes.LLM_PATH: APIS["CHAT_COMPLETION"]["ENDPOINT"],
-            **get_extra_attributes(),
+            **get_extra_attributes(),  # type: ignore
         }
 
-        attributes = LLMSpanAttributes(**span_attributes)
+        attributes = LLMSpanAttributes(**filter_valid_attributes(span_attributes))
 
         span = tracer.start_span(
-            APIS["CHAT_COMPLETION"]["METHOD"],
+            name=get_span_name(APIS["CHAT_COMPLETION"]["METHOD"]),
             kind=SpanKind.CLIENT,
             context=set_span_in_context(trace.get_current_span()),
         )
@@ -391,11 +387,9 @@ def async_chat_completions_create(original_method, version, tracer):
                         json.dumps((str(message))), kwargs.get("model")
                     )
 
-                if (
-                    kwargs.get("functions") is not None
-                    and kwargs.get("functions") != NOT_GIVEN
-                ):
-                    for function in kwargs.get("functions"):
+                functions = kwargs.get("functions")
+                if functions is not None and functions != NOT_GIVEN:
+                    for function in functions:
                         prompt_tokens += calculate_prompt_tokens(
                             json.dumps(function), kwargs.get("model")
                         )
@@ -406,9 +400,9 @@ def async_chat_completions_create(original_method, version, tracer):
                     prompt_tokens,
                     function_call=kwargs.get("functions") is not None,
                     tool_calls=kwargs.get("tools") is not None,
-                )
+                )  # type: ignore
             else:
-                _set_response_attributes(span, kwargs, result)
+                _set_response_attributes(span, result)
                 span.set_status(StatusCode.OK)
                 span.end()
                 return result
@@ -422,12 +416,17 @@ def async_chat_completions_create(original_method, version, tracer):
     return traced_method
 
 
-def embeddings_create(original_method, version, tracer):
+def embeddings_create(version: str, tracer: Tracer) -> Callable:
     """
     Wrap the `create` method of the `Embeddings` class to trace it.
     """
 
-    def traced_method(wrapped, instance, args, kwargs):
+    def traced_method(
+        wrapped: Callable,
+        instance: Any,
+        args: List[Any],
+        kwargs: EmbeddingsCreateKwargs,
+    ) -> Any:
         service_provider = SERVICE_PROVIDERS["OPENAI"]
 
         span_attributes = {
@@ -436,7 +435,7 @@ def embeddings_create(original_method, version, tracer):
             **get_llm_url(instance),
             SpanAttributes.LLM_PATH: APIS["EMBEDDINGS_CREATE"]["ENDPOINT"],
             SpanAttributes.LLM_REQUEST_DIMENSIONS: kwargs.get("dimensions"),
-            **get_extra_attributes(),
+            **get_extra_attributes(),  # type: ignore
         }
 
         encoding_format = kwargs.get("encoding_format")
@@ -452,10 +451,10 @@ def embeddings_create(original_method, version, tracer):
                 [kwargs.get("input", "")]
             )
 
-        attributes = LLMSpanAttributes(**span_attributes)
+        attributes = LLMSpanAttributes(**filter_valid_attributes(span_attributes))
 
         with tracer.start_as_current_span(
-            APIS["EMBEDDINGS_CREATE"]["METHOD"],
+            name=get_span_name(APIS["EMBEDDINGS_CREATE"]["METHOD"]),
             kind=SpanKind.CLIENT,
             context=set_span_in_context(trace.get_current_span()),
         ) as span:
@@ -479,12 +478,17 @@ def embeddings_create(original_method, version, tracer):
     return traced_method
 
 
-def async_embeddings_create(original_method, version, tracer):
+def async_embeddings_create(version: str, tracer: Tracer) -> Callable:
     """
     Wrap the `create` method of the `Embeddings` class to trace it.
     """
 
-    async def traced_method(wrapped, instance, args, kwargs):
+    async def traced_method(
+        wrapped: Callable,
+        instance: Any,
+        args: List[Any],
+        kwargs: EmbeddingsCreateKwargs,
+    ) -> Awaitable[Any]:
 
         service_provider = SERVICE_PROVIDERS["OPENAI"]
 
@@ -493,10 +497,10 @@ def async_embeddings_create(original_method, version, tracer):
             **get_llm_request_attributes(kwargs, operation_name="embed"),
             SpanAttributes.LLM_PATH: APIS["EMBEDDINGS_CREATE"]["ENDPOINT"],
             SpanAttributes.LLM_REQUEST_DIMENSIONS: kwargs.get("dimensions"),
-            **get_extra_attributes(),
+            **get_extra_attributes(),  # type: ignore
         }
 
-        attributes = LLMSpanAttributes(**span_attributes)
+        attributes = LLMSpanAttributes(**filter_valid_attributes(span_attributes))
 
         encoding_format = kwargs.get("encoding_format")
         if encoding_format is not None:
@@ -512,7 +516,7 @@ def async_embeddings_create(original_method, version, tracer):
             )
 
         with tracer.start_as_current_span(
-            APIS["EMBEDDINGS_CREATE"]["METHOD"],
+            name=get_span_name(APIS["EMBEDDINGS_CREATE"]["METHOD"]),
             kind=SpanKind.CLIENT,
             context=set_span_in_context(trace.get_current_span()),
         ) as span:
@@ -536,7 +540,7 @@ def async_embeddings_create(original_method, version, tracer):
     return traced_method
 
 
-def extract_content(choice):
+def extract_content(choice: Any) -> Union[str, List[Dict[str, Any]], Dict[str, Any]]:
     # Check if choice.message exists and has a content attribute
     if (
         hasattr(choice, "message")
@@ -581,13 +585,15 @@ def extract_content(choice):
 
 
 @silently_fail
-def _set_input_attributes(span, kwargs, attributes):
+def _set_input_attributes(
+    span: Span, kwargs: ChatCompletionsCreateKwargs, attributes: LLMSpanAttributes
+) -> None:
     tools = []
     for field, value in attributes.model_dump(by_alias=True).items():
         set_span_attribute(span, field, value)
-
-    if kwargs.get("functions") is not None and kwargs.get("functions") != NOT_GIVEN:
-        for function in kwargs.get("functions"):
+    functions = kwargs.get("functions")
+    if functions is not None and functions != NOT_GIVEN:
+        for function in functions:
             tools.append(json.dumps({"type": "function", "function": function}))
 
     if kwargs.get("tools") is not None and kwargs.get("tools") != NOT_GIVEN:
@@ -598,7 +604,7 @@ def _set_input_attributes(span, kwargs, attributes):
 
 
 @silently_fail
-def _set_response_attributes(span, kwargs, result):
+def _set_response_attributes(span: Span, result: ResultType) -> None:
     set_span_attribute(span, SpanAttributes.LLM_RESPONSE_MODEL, result.model)
     if hasattr(result, "choices") and result.choices is not None:
         responses = [
@@ -610,8 +616,8 @@ def _set_response_attributes(span, kwargs, result):
                 ),
                 "content": extract_content(choice),
                 **(
-                    {"content_filter_results": choice["content_filter_results"]}
-                    if "content_filter_results" in choice
+                    {"content_filter_results": choice.content_filter_results}
+                    if hasattr(choice, "content_filter_results")
                     else {}
                 ),
             }
